///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// (c) Copyright OCP-IP 2009-2011
// OCP-IP Confidential and Proprietary
//
//
//============================================================================
//      Project : OCP SLD WG
//       Author : James Aldis, Texas Instruments
//
//          $Id:
//
//  Description :  Testbench for TL1/TL0 adapters: precise aligned SRMD with
//                 varied burst types
//
// TL1 and TL0 initiators and targets are created with identical behaviour for
// the different abstraction levels.
// Six simulations are run, 4 with adapters and 2 without, for 0, 1 and 2
//   cascaded adapters.
// All sets of results should be the same, allowing for a few ps of timing
// delay in some cases.
//
// The TL1 OCP target in this testbench is not OCP compliant, in that it makes
// assumptions about burst alignment and TLM2 transaction alignment which
// initiators in general are not obliged to respect.
//                                                                           //
///////////////////////////////////////////////////////////////////////////////


#include "ocpip_adapters.h"
#include <iostream>
#include <queue>


using namespace std;
using namespace ocpip;
using namespace sc_core;
using namespace tlm;


// all OCPs in this testbench have the same configuration
ocp_parameters gen_ocp_config() {
  ocp_parameters p;
  // aligned INCR bursts length in {1,2,4,8,16,32}
  // also wrap, stream, xor and block
  // byte enables
  // no write response
  // non-posted and posted writes
  p.burst_aligned = true;
  p.burstlength = true;
  p.burstlength_wdth = 6;
  p.burstsinglereq = true;
  p.datahandshake = true;
  p.dataaccept = true;
  p.datalast = true;
  p.resplast = true;
  p.burstseq = true;
  p.burstseq_strm_enable = true;
  p.burstseq_wrap_enable = true;
  p.burstseq_xor_enable = true;
  p.burstseq_blck_enable = true;
  p.blockheight = true;
  p.blockheight_wdth = 5;
  p.blockstride = true;
  p.blockstride_wdth = 16;
  p.datarowlast = true;
  p.resprowlast = true;
  p.byteen = true;
  p.mdatabyteen = true;
  p.writenonpost_enable = true;
  // base config
  p.addr_wdth = 32;
  p.data_wdth = 32;
  p.mreset = true;
  return p;
}


// traits class for TL0 signal types
class SRMDSEQ_TYPES: public tl1_tl0::POD_DEFAULT_TL0_TYPES<> {
public:
  typedef unsigned MBURSTLENGTH_T;
  typedef bool MBURSTSINGLEREQ_T;
  typedef bool MDATAVALID_T;
  typedef bool SDATAACCEPT_T;
  typedef bool MDATALAST_T;
  typedef bool SRESPLAST_T;
  typedef unsigned MBURSTSEQ_T;
  typedef unsigned MBLOCKHEIGHT_T;
  typedef unsigned MBLOCKSTRIDE_T;
  typedef bool MDATAROWLAST_T;
  typedef bool SRESPROWLAST_T;
  typedef unsigned MBYTEEN_T;
  typedef unsigned MDATABYTEEN_T;
};


// simulation duration determined by these parameters - leave space for
// final burst to complete
#define MAX_ADDR 200000
#define MEM_SIZE 202000
// max block burst will be length 32 (128 bytes), height 4, stride 256
#include "tb_common.h"


class ocp_tl0_initiator: public sc_module {
SC_HAS_PROCESS(ocp_tl0_initiator);
public:
  ocp_tl0_initiator(sc_module_name n):
    sc_module(n),
    cycle(0),
    next_req_cycle(1),
    next_data_cycle(1),
    resps_rxd(0),
    resps_wanted(0),
    curr_addr(0),
    curr_burst_len(1),
    curr_burst_seq(BLCK),
    curr_cmd(RD),
    req_active(false),
    data_active(false),
    wr_data_len(0),
    rd_data_cnt(0),
    os(sc_module::name())
  {
    SC_METHOD(on_Clk_rising);
    dont_initialize();
    sensitive << Clk.pos();

    MReset_n.initialize(true);
    MCmd.initialize(0);
    MDataValid.initialize(false);

    fill_initiator_memory(memory);
  }

  sc_in<bool> Clk;
  sc_out<unsigned> MAddr;
  sc_out<unsigned> MCmd;
  sc_out<unsigned> MData;
  sc_in<unsigned> SData;
  sc_in<bool> SCmdAccept;
  sc_in<unsigned> SResp;
  sc_out<bool> MReset_n;

  sc_out<unsigned> MBurstLength;
  sc_out<bool> MBurstSingleReq;
  sc_out<bool> MDataValid;
  sc_out<bool> MDataLast;
  sc_in<bool> SDataAccept;
  sc_in<bool> SRespLast;
  sc_out<unsigned> MBurstSeq;
  sc_out<unsigned> MBlockHeight;
  sc_out<unsigned> MBlockStride;
  sc_out<bool> MDataRowLast;
  sc_in<bool> SRespRowLast;
  sc_out<unsigned> MByteEn;
  sc_out<unsigned> MDataByteEn;

private:
  enum burstseq {INCR = 0, WRAP = 2, XOR = 4, STRM = 5, BLCK = 7};
  enum command {WR = 1, RD = 2, WNP = 5};

  void on_Clk_rising() {
    if((curr_addr >= MAX_ADDR) && (resps_rxd == resps_wanted)) {
      stop_sim();
      return;
    }

    // request side
    if(req_active && SCmdAccept)
      os << cycle << " I: req accepted " << resps_wanted << endl;

    if(SCmdAccept || !req_active) {
      if((cycle >= next_req_cycle) && (curr_addr < MAX_ADDR)) {
        // SRMD bursts
        if(curr_burst_len == 1) {
          if((curr_addr & 127) == 0) curr_burst_len = 32;
        } else curr_burst_len /= 2;
        os << cycle << " I: SRMD req length " << curr_burst_len << endl;

        MBlockHeight = 1;
        MBlockStride = 0;
        unsigned length = curr_burst_len;
        unsigned offset = 0;
        unsigned next_addr = curr_addr + 4 * curr_burst_len;
        switch(curr_burst_seq) {
          case INCR:
            curr_burst_seq = WRAP;
            offset = 4 * (cycle % curr_burst_len);
            os << cycle << " I: WRAP req " << offset << endl;
            break;
          case WRAP:
            curr_burst_seq = XOR;
            offset = 4 * (cycle % curr_burst_len);
            os << cycle << " I: XOR req " << offset << endl;
            break;
          case XOR:
            curr_burst_seq = STRM;
            os << cycle << " I: STRM req" << endl;
            break;
          case STRM:
            curr_burst_seq = BLCK;
            length *= 4;
            next_addr += 3 * 256;
            MBlockHeight = 4;
            MBlockStride = 256;
            os << cycle << " I: BLCK req height 4 stride 256" << endl;
            break;
          case BLCK:
            curr_burst_seq = INCR;
            os << cycle << " I: INCR req" << endl;
            break;
        }

        // circle through WR, WNP and RD
        unsigned be = 0;
        switch(curr_cmd) {
          case RD:
            curr_cmd = WR;
            wr_addr.push(curr_addr);
            wr_len.push(length);
            wr_seq.push(curr_burst_seq);
            wr_off.push(offset);
            os << cycle << " I: W req sent @ " << curr_addr << endl;
            break;
          case WR:
            curr_cmd = WNP;
            rd_addr.push(-1);
            wr_addr.push(curr_addr);
            wr_len.push(length);
            wr_seq.push(curr_burst_seq);
            wr_off.push(offset);
            resps_wanted += 1;
            os << cycle << " I: WNP req sent @ " << curr_addr << endl;
            break;
          case WNP:
            curr_cmd = RD;
            be = (cycle & 16 ? 3 : 12);
            rd_addr.push(curr_addr);
            rd_seq.push(curr_burst_seq);
            rd_off.push(offset);
            rd_be.push(be);
            rd_len.push(length);
            resps_wanted += length;
            os << cycle << " I: R req sent @ " << curr_addr << ", BE = " << be << endl;
            break;
        }
        MCmd = curr_cmd;
        MAddr = curr_addr + offset;
        MBurstLength = curr_burst_len;
        MBurstSingleReq = true;
        MBurstSeq = curr_burst_seq;
        MByteEn = be;

        curr_addr = next_addr;
        req_active = true;

        // wait 8 x 1-65 cycles between requests roughly randomly
        unsigned sel = (cycle & 7) ^ ((cycle >> 3) & 7) ^ ((cycle >> 6) & 7);
        next_req_cycle = cycle + 8 * (1 + (1 << sel) / 2);
      } else {
        MCmd = 0;
        req_active = false;
      }
    }

    // data side
    if(data_active && SDataAccept)
      os << cycle << " I: data accepted " << resps_wanted << endl;

    if(SDataAccept || !data_active) {
      if((wr_addr.size() > 0) && (cycle >= next_data_cycle)) {
        MDataValid = true;
        unsigned be = (cycle & 16 ? 3 : 12);
        MDataByteEn = be;
        data_active = true;
        int &addr(wr_addr.front());
        int &len(wr_len.front());
        if(wr_data_len == 0) wr_data_len = len;
        unsigned seq = wr_seq.front();
        unsigned &off(wr_off.front());
        unsigned d = *reinterpret_cast<unsigned *>(memory + addr + off);
        MData = d;
        os << cycle << " I: W data: " << d << ", BE = " << be << endl;
        os << cycle << " I: W data sent @ " << addr << " + " << off << endl;
        os << cycle << " I: W length " << len << endl;
        if(1 == len) {
          wr_addr.pop();
          wr_len.pop();
          wr_seq.pop();
          wr_off.pop();
          wr_data_len = 0;
          MDataLast = true;
          MDataRowLast = true;
        } else {
          MDataRowLast = ((seq == BLCK) && (((off / 4 + 1) % (wr_data_len / 4)) == 0));
          switch(seq) {
            case BLCK:
            case INCR:
            case WRAP:
            case STRM:
              off = (off + 4) % (4 * wr_data_len);
              break;
            case XOR:
              // xor with 1,3,1,7,1,3,1,15,1,3,1,7,...
              off ^= 4 * ((32 - len) ^ (33 - len));
              break;
          }
          --len;
          MDataLast = false;
        }
        // wait 1,1,2,1,1,2,1,1,2,1,... cycles between data
        next_data_cycle = cycle + 1 + ((cycle / 2) & 1);
      } else {
        MDataValid = false;
        data_active = false;
      }
    }

    // response side
    if(SResp != 0) {
      resps_rxd++;
      int &addr(rd_addr.front());
      if(addr >= 0) {
        unsigned &off(rd_off.front());
        unsigned seq = rd_seq.front();
        unsigned be = rd_be.front();
        unsigned len = rd_len.front();
        unsigned sd = SData;
        if(be & 1) *(memory + addr + off) = sd; else sd &= 0xFFFFFF00;
        if(be & 2) *(memory + addr + off + 1) = (sd >> 8); else sd &= 0xFFFF00FF;
        if(be & 4) *(memory + addr + off + 2) = (sd >> 16); else sd &= 0xFF00FFFF;
        if(be & 8) *(memory + addr + off + 3) = (sd >> 24); else sd &= 0x00FFFFFF;
        os << cycle << " I: R data: " << sd << ", BE = " << be << endl;
        os << cycle << " I: R resp received @ " << addr << ", " << off << endl;
        if(SRespLast) {
          rd_addr.pop();
          rd_seq.pop();
          rd_off.pop();
          rd_be.pop();
          rd_len.pop();
          rd_data_cnt = 0;
          os << cycle << " I: R resp last" << endl;
        } else {
          switch(seq) {
            case INCR: case BLCK: case STRM:  off +=4; break;
            case WRAP:  off += 4; if(off == len * 4) off = 0; break;
            case XOR:  off ^= 4 * (rd_data_cnt ^ (rd_data_cnt + 1)); break;
          }
          ++rd_data_cnt;
        }
      } else {
        rd_addr.pop();
        os << cycle << " I: W resp received" << endl;
      }
    }

    cycle++;
  }

  char memory[MEM_SIZE];
  unsigned cycle;
  unsigned next_req_cycle, next_data_cycle, resps_rxd, resps_wanted;
  unsigned curr_addr, curr_burst_len, curr_burst_seq, curr_cmd;
  bool req_active, data_active, curr_is_read;
  queue<int> rd_addr, wr_addr, wr_len;
  queue<unsigned> rd_seq, wr_seq, rd_off, wr_off, rd_be, rd_len;
  unsigned wr_data_len;
  int rd_data_cnt;
  ofstream os;
};


class ocp_tl1_initiator: public sc_module {
SC_HAS_PROCESS(ocp_tl1_initiator);
public:
  ocp_tl1_initiator(sc_module_name n):
    sc_module(n),
    Clk("Clk"),
    ocp("ocp", this, &ocp_tl1_initiator::ocp_timing_update,
        ocp_master_socket_tl1<>::mm_txn_with_be()),
    cycle(0),
    next_req_cycle(1),
    next_data_cycle(1),
    resps_rxd(0),
    resps_wanted(0),
    curr_addr(0),
    curr_burst_len(1),
    curr_burst_seq(BLCK),
    curr_cmd(RD),
    wr_len(0),
    rd_len(0),
    req_active(false),
    data_active(false),
    scmdaccept(false),
    sdataaccept(false),
    sresp(0),
    os(sc_module::name()),
    srespT(SC_ZERO_TIME)
  {
    SC_METHOD(on_Clk_rising);
    dont_initialize();
    sensitive << Clk.pos();

    ocp.set_ocp_config(gen_ocp_config());
    ocp.activate_synchronization_protection();
    ocp.register_nb_transport_bw(this, &ocp_tl1_initiator::nb_transport_bw);

    fill_initiator_memory(memory);
  }

  sc_in<bool> Clk;
  ocp_master_socket_tl1<32> ocp;

  void end_of_elaboration() {set_master_timing();}

private:
  enum burstseq {INCR = 0, WRAP = 2, XOR = 4, STRM = 5, BLCK = 7};
  enum command {WR = 1, RD = 2, WNP = 5};

  struct ext: tlm_extension<ext> {
    void free() {delete this;}
    void copy_from(tlm_extension_base const &) {}
    tlm_extension_base *clone() const {return 0;}
    static ext &get(tlm_generic_payload &pl) {
      ext *e = pl.get_extension<ext>();
      if(e == 0) {e = new ext; pl.set_extension<ext>(e);}
      return *e;
    }
    unsigned addr, offset, seq;
    bool posted;
  };

  void on_Clk_rising() {
    if((curr_addr >= MAX_ADDR) && (resps_rxd == resps_wanted)) {
      stop_sim();
      return;
    }

    // request side
    if(req_active && scmdaccept)
      os << cycle << " I: req accepted " << resps_wanted << endl;

    if(scmdaccept || !req_active) {
      if((cycle >= next_req_cycle) && (curr_addr < MAX_ADDR)) {
        // SRMD bursts
        if(curr_burst_len == 1) {
          if((curr_addr & 127) == 0) curr_burst_len = 32;
        } else curr_burst_len /= 2;
        os << cycle << " I: SRMD req length " << curr_burst_len << endl;

        tlm_generic_payload *curr = ocp.get_transaction();
        leak_test(*curr);
        curr->set_streaming_width(16 * curr_burst_len);

        unsigned length = curr_burst_len;
        unsigned offset = 0;
        unsigned next_addr = curr_addr + 4 * curr_burst_len;
        burst_sequence *bs;
        switch(curr_burst_seq) {
          case INCR:
            curr_burst_seq = WRAP;
            offset = 4 * (cycle % curr_burst_len);
            ocp.get_extension<burst_sequence>(bs, *curr);
            bs->value.sequence = ocpip::WRAP;
            bs->value.xor_wrap_address = curr_addr + offset;
            ocp.validate_extension<burst_sequence>(*curr);
            os << cycle << " I: WRAP req " << offset << endl;
            break;
          case WRAP:
            curr_burst_seq = XOR;
            offset = 4 * (cycle % curr_burst_len);
            ocp.get_extension<burst_sequence>(bs, *curr);
            bs->value.sequence = ocpip::XOR;
            bs->value.xor_wrap_address = curr_addr + offset;
            ocp.validate_extension<burst_sequence>(*curr);
            os << cycle << " I: XOR req " << offset << endl;
            break;
          case XOR:
            curr_burst_seq = STRM;
            curr->set_streaming_width(4);
            if(length == 1) {
              // stream burst of length 1 is functionally equivalent to INCR
              // but may be treated differently by interconnects and targets
              ocp.get_extension<burst_sequence>(bs, *curr);
              bs->value.sequence = ocpip::STRM;
              ocp.validate_extension<burst_sequence>(*curr);
            }
            else ocp.invalidate_extension<burst_sequence>(*curr);
            os << cycle << " I: STRM req" << endl;
            break;
          case STRM:
            curr_burst_seq = BLCK;
            next_addr += 3 * 256;
            length *= 4;
            ocp.get_extension<burst_sequence>(bs, *curr);
            bs->value.sequence = ocpip::BLCK;
            bs->value.block_height = 4;
            bs->value.block_stride = 256;
            bs->value.blck_row_length_in_bytes = 4 * curr_burst_len;
            ocp.validate_extension<burst_sequence>(*curr);
            os << cycle << " I: BLCK req height 4 stride 256" << endl;
            break;
          case BLCK:
            curr_burst_seq = INCR;
            ocp.invalidate_extension<burst_sequence>(*curr);
            os << cycle << " I: INCR req" << endl;
            break;
        }
        ext &e(ext::get(*curr));
        e.addr = curr_addr;
        e.offset = offset;
        e.seq = curr_burst_seq;

        // circle through WR, WNP and RD
        ocp.reserve_be_size(*curr, 4 * length);
        switch(curr_cmd) {
          case RD:
            curr_cmd = WR;
            curr->set_write();
            ocp.validate_extension<posted>(*curr);
            e.posted = true;
            wr_addr.push(curr);
            os << cycle << " I: W req sent @ " << curr_addr << endl;
            break;
          case WR:
            curr_cmd = WNP;
            curr->set_write();
            ocp.invalidate_extension<posted>(*curr);
            e.posted = false;
            wr_addr.push(curr);
            resps_wanted += 1;
            os << cycle << " I: WNP req sent @ " << curr_addr << endl;
            break;
          case WNP:
            curr_cmd = RD;
            curr->set_read();
            ocp.reserve_be_size(*curr, 4);
            unsigned be = (cycle & 16 ? 3 : 12);
            unsigned char *bep = curr->get_byte_enable_ptr();
            bep[0] = (be & 1 ? TLM_BYTE_ENABLED : TLM_BYTE_DISABLED);
            bep[1] = (be & 2 ? TLM_BYTE_ENABLED : TLM_BYTE_DISABLED);
            bep[2] = (be & 4 ? TLM_BYTE_ENABLED : TLM_BYTE_DISABLED);
            bep[3] = (be & 8 ? TLM_BYTE_ENABLED : TLM_BYTE_DISABLED);
            resps_wanted += length;
            os << cycle << " I: R req sent @ " << curr_addr << ", BE = " << be << endl;
            break;
        }

        ocp.validate_extension<srmd>(*curr);
        curr->set_response_status(TLM_INCOMPLETE_RESPONSE);
        curr->set_address(curr_addr);
        curr->set_data_ptr((unsigned char*)memory + curr_addr);
        curr->set_data_length(4 * length);

        tlm_phase ph = BEGIN_REQ;
        sc_time t = SC_ZERO_TIME;
        if(ocp->nb_transport_fw(*curr, ph, t) == TLM_UPDATED) {
          sc_assert(ph == END_REQ);
          scmdaccept = true;
        } else {
          scmdaccept = false;
        }

        curr_addr = next_addr;
        req_active = true;

        // wait 8 x 1-65 cycles between requests roughly randomly
        unsigned sel = (cycle & 7) ^ ((cycle >> 3) & 7) ^ ((cycle >> 6) & 7);
        next_req_cycle = cycle + 8 * (1 + (1 << sel) / 2);
      } else {
        req_active = false;
      }
    }

    // data side
    if(data_active && sdataaccept)
      os << cycle << " I: data accepted " << resps_wanted << endl;

    if(sdataaccept || !data_active) {
      if((wr_addr.size() > 0) && (cycle >= next_data_cycle)) {
        tlm_generic_payload *curr = wr_addr.front();
        ext &e(ext::get(*curr));

        unsigned be = (cycle & 16 ? 3 : 12);
        unsigned char *bep = curr->get_byte_enable_ptr() + e.offset;
        bep[0] = (be & 1 ? TLM_BYTE_ENABLED : TLM_BYTE_DISABLED);
        bep[1] = (be & 2 ? TLM_BYTE_ENABLED : TLM_BYTE_DISABLED);
        bep[2] = (be & 4 ? TLM_BYTE_ENABLED : TLM_BYTE_DISABLED);
        bep[3] = (be & 8 ? TLM_BYTE_ENABLED : TLM_BYTE_DISABLED);

        unsigned len = curr->get_data_length() / 4;
        unsigned d = *reinterpret_cast<unsigned *>(memory + e.addr + e.offset);
        os << cycle << " I: W data: " << d << ", BE = " << be << endl;
        os << cycle << " I: W data sent @ " << e.addr << " + " << e.offset << endl;
        os << cycle << " I: W length " << len - wr_len << endl;

        tlm_phase ph = BEGIN_DATA;
        sc_time t = SC_ZERO_TIME;
        if(ocp->nb_transport_fw(*curr, ph, t) == TLM_UPDATED) {
          sc_assert(ph == END_DATA);
          sdataaccept = true;
        } else {
          sdataaccept = false;
        }
        data_active = true;

        if(len == ++wr_len) {
          wr_addr.pop();
          wr_len = 0;
          if(e.posted) ocp.release_transaction(curr);
        } else switch(e.seq) {
          case WRAP: e.offset = (e.offset + 4) % (4 * len); break;
          case XOR: e.offset ^= 4 * (wr_len ^ (wr_len - 1)); break;
          case BLCK: case INCR: case STRM: e.offset += 4; break;
        }

        // wait 1,1,2,1,1,2,1,1,2,1,... cycles between data
        next_data_cycle = cycle + 1 + ((cycle / 2) & 1);
      } else {
        data_active = false;
      }
    }

    // response side
    if(sresp != 0) {
      resps_rxd++;
      if(sresp->is_read()) {
        ext &e(ext::get(*sresp));
        unsigned len = sresp->get_data_length() / 4;
        unsigned d = *reinterpret_cast<unsigned *>(memory + e.addr + e.offset);
        unsigned be = 0;
        unsigned char *bep = sresp->get_byte_enable_ptr();
        // byte enable len is 4 for RD in this initiator
        if(bep[0] == TLM_BYTE_ENABLED) be |= 1; else d &= 0xFFFFFF00;
        if(bep[1] == TLM_BYTE_ENABLED) be |= 2; else d &= 0xFFFF00FF;
        if(bep[2] == TLM_BYTE_ENABLED) be |= 4; else d &= 0xFF00FFFF;
        if(bep[3] == TLM_BYTE_ENABLED) be |= 8; else d &= 0x00FFFFFF;
        os << cycle << " I: R data: " <<  d << ", BE = " << be << endl;
        os << cycle << " I: R resp received @ " << e.addr << ", " << e.offset << endl;
        if(++rd_len == len) {
          rd_len = 0;
          ocp.release_transaction(sresp);
          os << cycle << " I: R resp last" << endl;
        } else switch(e.seq) {
          case WRAP: e.offset = (e.offset + 4) % (4 * len); break;
          case XOR: e.offset ^= 4 * (rd_len ^ (rd_len - 1)); break;
          case STRM: case BLCK: case INCR: e.offset += 4; break;
        }
      } else {
        os << cycle << " I: W resp received" << endl;
        ocp.release_transaction(sresp);
      }
      sresp = 0;
    }

    cycle++;
  }

  tlm_sync_enum nb_transport_bw(
    tlm_generic_payload &pl, tlm_phase &ph, sc_time &ta)
  {
    if(ph == END_REQ) {
      scmdaccept = true;
      return TLM_ACCEPTED;
    }
    if(ph == END_DATA) {
      sdataaccept = true;
      return TLM_ACCEPTED;
    }
    sc_assert(ph == BEGIN_RESP);
    sresp = &pl;
    ph = END_RESP;
    return TLM_UPDATED;
  }

  char memory[MEM_SIZE];
  unsigned cycle;
  unsigned next_req_cycle, next_data_cycle, resps_rxd, resps_wanted;
  unsigned curr_addr, curr_burst_len, curr_burst_seq, curr_cmd, wr_len, rd_len;
  bool req_active, data_active;
  bool scmdaccept, sdataaccept;
  tlm_generic_payload *sresp;
  queue<tlm_generic_payload *> wr_addr;
  ofstream os;
  sc_time srespT;

  // not sure if this is redundant when there is no resp-accept in the OCP
  // configuration.  It still exists in the TLM2
  void ocp_timing_update(ocp_tl1_slave_timing times) {
    if(srespT != times.ResponseGrpStartTime) {
      srespT = times.ResponseGrpStartTime;
      set_master_timing();
    }
  }

  void set_master_timing() {
    ocp_tl1_master_timing mytimes;
    mytimes.MRespAcceptStartTime = srespT + sc_get_time_resolution();
    ocp.set_master_timing(mytimes);
  }
};


class ocp_tl0_target: public sc_module {
SC_HAS_PROCESS(ocp_tl0_target);
public:
  ocp_tl0_target(sc_module_name n):
    sc_module(n),
    cycle(0),
    resp_cnt(0),
    os(sc_module::name())
  {
    SC_METHOD(on_Clk_rising);
    dont_initialize();
    sensitive << Clk.pos() << MReset_n.neg();

    SResp.initialize(0);
    SCmdAccept.initialize(false);
    SDataAccept.initialize(false);

    fill_target_memory(memory);
  }

  sc_in<bool> Clk;
  sc_in<unsigned> MAddr;
  sc_in<unsigned> MCmd;
  sc_in<unsigned> MData;
  sc_out<unsigned> SData;
  sc_out<bool> SCmdAccept;
  sc_out<unsigned> SResp;
  sc_in<bool> MReset_n;

  sc_in<unsigned> MBurstLength;
  sc_in<bool> MBurstSingleReq;
  sc_in<bool> MDataValid;
  sc_in<bool> MDataLast;
  sc_out<bool> SDataAccept;
  sc_out<bool> SRespLast;
  sc_in<unsigned> MBurstSeq;
  sc_in<unsigned> MBlockHeight;
  sc_in<unsigned> MBlockStride;
  sc_in<bool> MDataRowLast;
  sc_out<bool> SRespRowLast;
  sc_in<unsigned> MByteEn;
  sc_in<unsigned> MDataByteEn;

private:
  enum burstseq {INCR = 0, WRAP = 2, XOR = 4, STRM = 5, BLCK = 7};
  enum command {WR = 1, RD = 2, WNP = 5};

  void on_Clk_rising() {
    SResp = 0;
    if(!MReset_n) {
      // reset all state
      while(rd_addr.size() > 0) rd_addr.pop();
      while(rd_len.size() > 0) rd_len.pop();
      while(rd_seq.size() > 0) rd_seq.pop();
      while(rd_height.size() > 0) rd_height.pop();
      while(rd_stride.size() > 0) rd_stride.pop();
      while(wr_addr.size() > 0) wr_addr.pop();
      while(wr_cmd.size() > 0) wr_cmd.pop();
      while(wr_seq.size() > 0) wr_seq.pop();
      while(wr_len.size() > 0) wr_len.pop();
      while(wr_stride.size() > 0) wr_stride.pop();
      while(rd_cycle.size() > 0) rd_cycle.pop();
      while(wr_cycle.size() > 0) wr_cycle.pop();
    } else {
      // response side
      if(rd_addr.size() > 0) {
        bool is_read = (rd_addr.front() >= 0);
        if(is_read && (rd_cycle.size() > 0) && (cycle >= rd_cycle.front())) {
          // do a read response
          SResp = 1;
          int &addr(rd_addr.front());
          int len = rd_len.front();
          unsigned d = *reinterpret_cast<unsigned *>(memory + addr);
          SData = d;
          os << cycle << " T: R data: " <<  d << endl;
          os << cycle << " T: R resp sent @ " << addr << endl;
          if(len == resp_cnt + 1) {
            SRespRowLast = true;
            if(--rd_height.front() == 0) {
              SRespLast = true;
              rd_addr.pop();
              rd_len.pop();
              rd_seq.pop();
              rd_height.pop();
              rd_stride.pop();
            } else {
              SRespLast = false;
              addr += 4 - len * 4 + rd_stride.front();
            }
            resp_cnt = 0;
          } else {
            SRespLast = false;
            SRespRowLast = false;
            switch(rd_seq.front()) {
              case INCR: case BLCK: addr += 4; break;
              case WRAP: addr += 4; if(4*len*(addr/(4*len)) == addr) addr -= 4*len; break;
              case XOR: addr ^= 4 * (resp_cnt ^ (resp_cnt+1)); break;
              case STRM: break;
            }
            ++resp_cnt;
          }
          rd_cycle.pop();
        }
        if((!is_read) && (wr_cycle.size() > 0) && (cycle >= wr_cycle.front())) {
          // do a write response
          SResp = 1;
          os << cycle << " T: W resp sent" << endl;
          SRespLast = true;
          SRespRowLast = true;
          rd_addr.pop();
          wr_cycle.pop();
        }
      }

      // request side
      // accept in cycles 5,10,15,... and impose 4 cycle minimum latency
      if(SCmdAccept) switch(MCmd) {
        case 0:
          break;
        case WR:
          sc_assert(MBurstSingleReq);
          wr_addr.push(MAddr);
          wr_cmd.push(WR);
          wr_seq.push(MBurstSeq);
          wr_len.push(MBurstLength);
          wr_stride.push(MBlockStride);
          os << cycle << " T: W req received @ " << MAddr << endl;
          os << cycle << " T: W req length " << MBurstLength << endl;
          os << cycle << " T: W req seq " << MBurstSeq << endl;
          break;
        case RD:
          sc_assert(MBurstSingleReq);
          rd_addr.push(MAddr);
          rd_len.push(MBurstLength);
          rd_seq.push(MBurstSeq);
          rd_height.push(MBurstSeq == BLCK ? MBlockHeight : 1);
          rd_stride.push(MBlockStride);
          os << cycle << " T: R req received @ " << MAddr << endl;
          os << cycle << " T: R req length " << MBurstLength << endl;
          os << cycle << " T: R req seq " << MBurstSeq << endl;
          for(unsigned i=0; i<MBurstLength*(MBurstSeq==BLCK?MBlockHeight:1); i++)
            rd_cycle.push((i/4)*4 + cycle + 4);
          break;
        case WNP:
          sc_assert(MBurstSingleReq);
          rd_addr.push(-1);
          wr_addr.push(MAddr);
          wr_cmd.push(WNP);
          wr_seq.push(MBurstSeq);
          wr_len.push(MBurstLength);
          wr_stride.push(MBlockStride);
          os << cycle << " T: WNP req received @ " << MAddr << endl;
          os << cycle << " T: WNP req length " << MBurstLength << endl;
          os << cycle << " T: WNP req seq " << MBurstSeq << endl;
          break;
      }
      bool new_scmdaccept = ((cycle % 5) == 0) || ((cycle % 17) == 0);
      SCmdAccept = new_scmdaccept;

      // data side
      // accept in cycles 3,6,9,12,...
      // but do not accept data before accepting request
      if(SDataAccept && MDataValid) {
        int &addr(wr_addr.front());
        unsigned d = MData;
        unsigned be = MDataByteEn;
        if(be & 1) *(memory + addr) = d; else d &= 0xFFFFFF00;
        if(be & 2) *(memory + addr + 1) = (d >> 8); else d &= 0xFFFF00FF;
        if(be & 4) *(memory + addr + 2) = (d >> 16); else d &= 0xFF00FFFF;
        if(be & 8) *(memory + addr + 3) = (d >> 24); else d &= 0x00FFFFFF;
        os << cycle << " T: W data: " << d << ", BE = " << be << endl;
        if(MDataLast) {
          if(wr_cmd.front() == WNP) wr_cycle.push(cycle + 4);
          wr_addr.pop();
          wr_cmd.pop();
          wr_seq.pop();
          wr_len.pop();
          wr_stride.pop();
          os << cycle << " T: W data last" << endl;
        } else {
          int &len(wr_len.front());
          switch(wr_seq.front()) {
            case INCR: addr += 4; break;
            case WRAP: addr += 4; if(4*len*(addr/(4*len)) == addr) addr -= 4 * len; break;
            case XOR: addr ^= 4 * ((32 - len) ^ (32 - (--len))); break;
            case STRM: break;
            case BLCK: addr += 4; if(MDataRowLast) addr += wr_stride.front() - 4 * len; break;
          }
        }
      }
      SDataAccept = (((cycle % 3) == 0) || ((cycle % 13) == 0))
        && ((wr_addr.size() > 0) || new_scmdaccept);

      cycle++;
    }
  }

  unsigned cycle;
  int resp_cnt;
  char memory[MEM_SIZE];
  queue<int> rd_addr, rd_len, wr_addr, wr_len;
  queue<unsigned> rd_cycle, wr_cycle;
  queue<unsigned> wr_cmd, wr_seq, wr_stride, rd_seq, rd_height, rd_stride;
  ofstream os;
};


//FIXME NEED TO CONSIDER burst_length AND blck_width_in_bytes AND streaming_width INSTEAD OF ASSUMPTIONS
class ocp_tl1_target: public sc_module {
SC_HAS_PROCESS(ocp_tl1_target);
public:
  ocp_tl1_target(sc_module_name n):
    sc_module(n),
    Clk("Clk"),
    ocp("ocp", this, &ocp_tl1_target::ocp_timing_update),
    cycle(0),
    in_reset(false),
    scmdaccept(false),
    sdataaccept(false),
    mcmd(0),
    mdatahs(0),
    wr_len(0),
    rd_len(0),
    wr_addr_size(0),
    os(sc_module::name()),
    cmdT(SC_ZERO_TIME),
    dataT(SC_ZERO_TIME)
  {
    SC_METHOD(on_Clk_rising);
    dont_initialize();
    sensitive << Clk.pos();

    ocp.set_ocp_config(gen_ocp_config());
    ocp.activate_synchronization_protection();
    ocp.register_nb_transport_fw(this, &ocp_tl1_target::nb_transport_fw);

    fill_target_memory(memory);
    enabled[0] = enabled[1] = enabled[2] = enabled[3] = TLM_BYTE_ENABLED;
  }

  sc_in<bool> Clk;
  ocp_slave_socket_tl1<32> ocp;

  void end_of_elaboration() {set_slave_timing();}

private:
  enum burstseq {INCR = 0, WRAP = 2, XOR = 4, STRM = 5, BLCK = 7};
  enum command {WR = 1, RD = 2, WNP = 5};

  struct ext: tlm_extension<ext> {
    void free() {delete this;}
    void copy_from(tlm_extension_base const &) {}
    tlm_extension_base *clone() const {return 0;}
    static ext &get(tlm_generic_payload &pl) {
      ext *e = pl.get_extension<ext>();
      if(e == 0) {e = new ext; pl.set_extension<ext>(e);}
      return *e;
    }
    unsigned addr, offset, seq, height, stride;
    bool posted;
  };

  void on_Clk_rising() {
    if(in_reset) return;

    // response side
    if(rd_addr.size() > 0) {
      tlm_generic_payload *pl = rd_addr.front();
      if(pl->is_read() && (rd_cycle.size() > 0) && (cycle >= rd_cycle.front()))
      {
        // do a read response
        ext &e(ext::get(*pl));
        unsigned len = pl->get_data_length() / 4;
        unsigned width = len / 4;
        unsigned row = (e.seq != BLCK ? 0 : rd_len / width);
        unsigned strm = (e.seq != STRM ? 0 : rd_len * 4);

        unsigned char *local_addr = (unsigned char *)memory + e.addr + e.offset + row * e.stride;
        os << cycle << " T: R data: " <<  *reinterpret_cast<unsigned*>(local_addr) << endl;
        os << cycle << " T: R resp sent @ " << e.addr + e.offset + row * e.stride << endl;
        unsigned char *init_addr = pl->get_data_ptr() + e.offset + row * width * 4 + strm;
        unsigned char *be_addr = pl->get_byte_enable_ptr();
        if(be_addr == 0) be_addr = enabled;
        else be_addr += ((e.offset + row * width * 4 + strm) % pl->get_byte_enable_length());
        for(unsigned i=0; i<4; i++) {
          if(*be_addr++ == TLM_BYTE_ENABLED) *init_addr = *local_addr;  // FIXME assumes be-length is a multiple of 4
          ++init_addr;
          ++local_addr;
        }

        pl->set_response_status(TLM_OK_RESPONSE);
        tlm_phase ph = BEGIN_RESP;
        sc_time t = SC_ZERO_TIME;
        ocp->nb_transport_bw(*pl, ph, t);
        if(++rd_len == len) {
          rd_addr.pop();
          pl->release();
          rd_len = 0;
        } else switch(e.seq) {
          case WRAP: e.offset = (e.offset + 4) % (4 * len); break;
          case XOR: e.offset ^= 4 * (rd_len ^ (rd_len - 1)); break;
          case BLCK: e.offset = (rd_len % width) == 0 ? 0 : e.offset + 4; break;
          case INCR: e.offset += 4; break;
          case STRM: break;
        }
        rd_cycle.pop();
      }
      if(pl->is_write() && (wr_cycle.size() > 0) && (cycle >= wr_cycle.front()))
      {
        // do a write response
        pl->set_response_status(TLM_OK_RESPONSE);
        os << cycle << " T: W resp sent" << endl;
        tlm_phase ph = BEGIN_RESP;
        sc_time t = SC_ZERO_TIME;
        ocp->nb_transport_bw(*pl, ph, t);
        rd_addr.pop();
        pl->release();
        wr_cycle.pop();
      }
    }

    // request side
    // accept in cycles 5,10,15,... and impose 4 cycle minimum latency
    if(scmdaccept && (mcmd != 0)) {
      sc_assert(ocp.get_extension<srmd>(*mcmd) != 0);
      ext &e(ext::get(*mcmd));
      unsigned len = mcmd->get_data_length() / 4;

      burst_sequence *bs;
      e.offset = 0;
      e.addr = mcmd->get_address();
      e.height = 1;
      if(ocp.get_extension<burst_sequence>(bs, *mcmd)) {
        e.seq = bs->value.sequence;
        if(e.seq == BLCK) e.height = bs->value.block_height;
        e.stride = bs->value.block_stride;
        if((e.seq == XOR) || (e.seq == WRAP)) {
          e.offset = bs->value.xor_wrap_address - e.addr;
        }
      } else {
        e.seq = (mcmd->get_streaming_width() >= 4 * len ? INCR : STRM);
      }

      if(mcmd->is_write()) {
        if(ocp.get_extension<posted>(*mcmd) != 0) {
          os << cycle << " T: W req received @ " << e.addr + e.offset << endl;
          os << cycle << " T: W req length " << len / e.height << endl;
          os << cycle << " T: W req seq " << e.seq << endl;
          e.posted = true;
        } else {
          os << cycle << " T: WNP req received @ " << e.addr + e.offset << endl;
          os << cycle << " T: WNP req length " << len / e.height << endl;
          os << cycle << " T: WNP req seq " << e.seq << endl;
          e.posted = false;
        }
        ++wr_addr_size;
      } else {
        os << cycle << " T: R req received @ " << e.addr + e.offset << endl;
        os << cycle << " T: R req length " << len / e.height << endl;
        os << cycle << " T: R req seq " << e.seq << endl;
        for(unsigned i=0; i<len; i++) rd_cycle.push((i/4)*4 + cycle + 4);
      }
      mcmd = 0;
    }
    scmdaccept = ((cycle % 5) == 0) || ((cycle % 17) == 0);
    if(scmdaccept && (mcmd != 0)) {
      tlm_phase ph = END_REQ;
      sc_time t = SC_ZERO_TIME;
      ocp->nb_transport_bw(*mcmd, ph, t);
    }

    // data side
    // accept in cycles 3,6,9,12,...
    // but do not accept data before accepting request
    if(sdataaccept && (mdatahs != 0)) {
      ext &e(ext::get(*mdatahs));
      unsigned len = mdatahs->get_data_length() / 4;
      unsigned width = len / 4;
      unsigned row = (e.seq != BLCK ? 0 : wr_len / width);
      unsigned strm = (e.seq != STRM ? 0 : 4 * wr_len);

      unsigned char *local_addr = (unsigned char *)memory + e.addr + e.offset + row * e.stride;
      unsigned char *init_addr = mdatahs->get_data_ptr() + e.offset + row * width * 4 + strm;
      unsigned char *bep = mdatahs->get_byte_enable_ptr();
      if(bep == 0) bep = enabled;
      else bep += ((e.offset + row * width * 4 + strm) % mdatahs->get_byte_enable_length());
      unsigned d = *reinterpret_cast<unsigned *>(init_addr);
      unsigned be = (bep[0] & 1) | (bep[1] & 2) | (bep[2] & 4) | (bep[3] & 8);
      for(unsigned i=0; i<4; i++) {
        if(*bep++ == TLM_BYTE_ENABLED) *local_addr = *init_addr;  // assumes byte enable len is multiple of 4
        else d &= ~(0xFF << (8*i));
        ++init_addr;
        ++local_addr;
      }
      os << cycle << " T: W data: " << d << ", BE = " << be << endl;

      if(++wr_len == len) {
        if(e.posted) mdatahs->release();
        else wr_cycle.push(cycle + 4);
        wr_len = 0;
        --wr_addr_size;
        os << cycle << " T: W data last" << endl;
      } else switch(e.seq) {
        case INCR: case WRAP: e.offset = (e.offset + 4) % (4 * len); break;
        case XOR: e.offset ^= 4 * (wr_len ^ (wr_len - 1)); break;
        case STRM: break;
        case BLCK: e.offset = ((wr_len % width) == 0 ? 0 : e.offset + 4); break;
      }
      mdatahs = 0;
    }
    sdataaccept = (((cycle % 3) == 0) || ((cycle % 13) == 0))
        && ((wr_addr_size > 0) || scmdaccept);
    if(sdataaccept && (mdatahs != 0)) {
      tlm_phase ph = END_DATA;
      sc_time t = SC_ZERO_TIME;
      ocp->nb_transport_bw(*mdatahs, ph, t);
    }

    cycle++;
  }

  tlm_sync_enum nb_transport_fw(
    tlm_generic_payload &pl, tlm_phase &ph, sc_time &ta)
  {
    if(ph == END_RESET) {
      in_reset = false;
      while(rd_addr.size() > 0) {
        rd_addr.front()->release();
        rd_addr.pop();
      }
      return TLM_ACCEPTED;
    }
    if(ph == BEGIN_RESET) {
      in_reset = true;
      mcmd = 0;
      ocp.reset();
      while(rd_cycle.size() > 0) rd_cycle.pop();
      while(wr_cycle.size() > 0) wr_cycle.pop();
      return TLM_ACCEPTED;
    }
    if(ph == BEGIN_REQ) {
      leak_test(pl);
      mcmd = &pl;
      if(pl.is_read() || (ocp.get_extension<posted>(pl) == 0)) rd_addr.push(&pl);
      pl.acquire();
      if(scmdaccept) {
        ph = END_REQ;
        return TLM_UPDATED;
      } else {
        return TLM_ACCEPTED;
      }
    }
    if(ph == BEGIN_DATA) {
      mdatahs = &pl;
      if(sdataaccept) {
        ph = END_DATA;
        return TLM_UPDATED;
      } else {
        return TLM_ACCEPTED;
      }
    }
    sc_assert(ph == END_RESP);
    return TLM_ACCEPTED;
  }

  unsigned cycle;
  bool in_reset;
  bool scmdaccept, sdataaccept;
  tlm_generic_payload *mcmd, *mdatahs;
  unsigned wr_len, rd_len, wr_addr_size;
  char memory[MEM_SIZE];
  unsigned char enabled[4];
  queue<tlm_generic_payload *> rd_addr;
  queue<unsigned> rd_cycle, wr_cycle;
  ofstream os;
  sc_time cmdT, dataT;

  void ocp_timing_update(ocp_tl1_master_timing times) {
    if((cmdT != times.RequestGrpStartTime) ||
       (dataT != times.DataHSGrpStartTime)) {
      cmdT = times.RequestGrpStartTime;
      dataT = times.DataHSGrpStartTime;
      set_slave_timing();
    }
  }

  void set_slave_timing() {
    ocp_tl1_slave_timing mytimes;
    mytimes.SCmdAcceptStartTime = cmdT + sc_get_time_resolution();
    mytimes.SDataAcceptStartTime = dataT + sc_get_time_resolution();
    ocp.set_slave_timing(mytimes);
  }
};


class testbench00: public sc_module {
public:
  testbench00(sc_module_name n):
    sc_module(n),
    Clk("Clk"),
    initiator("initiator"),
    target("target")
  {
    // clock
    initiator.Clk(Clk);
    target.Clk(Clk);

    // TL0 interface
    initiator.MAddr(MAddr);
    initiator.MCmd(MCmd);
    initiator.MData(MData);
    initiator.SData(SData);
    initiator.SCmdAccept(SCmdAccept);
    initiator.SResp(SResp);
    initiator.MReset_n(MReset_n);
    target.MAddr(MAddr);
    target.MCmd(MCmd);
    target.MData(MData);
    target.SData(SData);
    target.SCmdAccept(SCmdAccept);
    target.SResp(SResp);
    target.MReset_n(MReset_n);

    initiator.MBurstLength(MBurstLength);
    initiator.MBurstSingleReq(MBurstSingleReq);
    initiator.MDataValid(MDataValid);
    initiator.MDataLast(MDataLast);
    initiator.SDataAccept(SDataAccept);
    initiator.SRespLast(SRespLast);
    initiator.MBurstSeq(MBurstSeq);
    initiator.MBlockHeight(MBlockHeight);
    initiator.MBlockStride(MBlockStride);
    initiator.MDataRowLast(MDataRowLast);
    initiator.SRespRowLast(SRespRowLast);
    initiator.MByteEn(MByteEn);
    initiator.MDataByteEn(MDataByteEn);
    target.MBurstLength(MBurstLength);
    target.MBurstSingleReq(MBurstSingleReq);
    target.MDataValid(MDataValid);
    target.MDataLast(MDataLast);
    target.SDataAccept(SDataAccept);
    target.SRespLast(SRespLast);
    target.MBurstSeq(MBurstSeq);
    target.MBlockHeight(MBlockHeight);
    target.MBlockStride(MBlockStride);
    target.MDataRowLast(MDataRowLast);
    target.SRespRowLast(SRespRowLast);
    target.MByteEn(MByteEn);
    target.MDataByteEn(MDataByteEn);
  }

  sc_in<bool> Clk;

private:
  ocp_tl0_initiator initiator;
  ocp_tl0_target target;

  sc_signal<unsigned> MAddr;
  sc_signal<unsigned> MCmd;
  sc_signal<unsigned> MData;
  sc_signal<unsigned> SData;
  sc_signal<bool> SCmdAccept;
  sc_signal<unsigned> SResp;
  sc_signal<bool> MReset_n;

  sc_signal<unsigned> MBurstLength;
  sc_signal<bool> MBurstSingleReq;
  sc_signal<bool> MDataValid;
  sc_signal<bool> MDataLast;
  sc_signal<bool> SDataAccept;
  sc_signal<bool> SRespLast;
  sc_signal<unsigned> MBurstSeq;
  sc_signal<unsigned> MBlockHeight;
  sc_signal<unsigned> MBlockStride;
  sc_signal<bool> MDataRowLast;
  sc_signal<bool> SRespRowLast;
  sc_signal<unsigned> MByteEn;
  sc_signal<unsigned> MDataByteEn;
};


class testbench01: public sc_module {
public:
  testbench01(sc_module_name n):
    sc_module(n),
    Clk("Clk"),
    initiator("initiator"),
    target("target"),
    adapter("adapter")
  {
    // OCP configuration
    adapter.set_ocp_config(gen_ocp_config());

    // clock
    initiator.Clk(Clk);
    target.Clk(Clk);
    adapter.Clk(Clk);

    // TL0 interface
    initiator.MAddr(MAddr);
    initiator.MCmd(MCmd);
    initiator.MData(MData);
    initiator.SData(SData);
    initiator.SCmdAccept(SCmdAccept);
    initiator.SResp(SResp);
    initiator.MReset_n(MReset_n);
    adapter.MAddr(MAddr);
    adapter.MCmd(MCmd);
    adapter.MData(MData);
    adapter.SData(SData);
    adapter.SCmdAccept(SCmdAccept);
    adapter.SResp(SResp);
    adapter.MReset_n(MReset_n);

    initiator.MBurstLength(MBurstLength);
    initiator.MBurstSingleReq(MBurstSingleReq);
    initiator.MDataValid(MDataValid);
    initiator.MDataLast(MDataLast);
    initiator.SDataAccept(SDataAccept);
    initiator.SRespLast(SRespLast);
    initiator.MBurstSeq(MBurstSeq);
    initiator.MBlockHeight(MBlockHeight);
    initiator.MBlockStride(MBlockStride);
    initiator.MDataRowLast(MDataRowLast);
    initiator.SRespRowLast(SRespRowLast);
    initiator.MByteEn(MByteEn);
    initiator.MDataByteEn(MDataByteEn);
    adapter.MBurstLength(MBurstLength);
    adapter.MBurstSingleReq(MBurstSingleReq);
    adapter.MDataValid(MDataValid);
    adapter.MDataLast(MDataLast);
    adapter.SDataAccept(SDataAccept);
    adapter.SRespLast(SRespLast);
    adapter.MBurstSeq(MBurstSeq);
    adapter.MBlockHeight(MBlockHeight);
    adapter.MBlockStride(MBlockStride);
    adapter.MDataRowLast(MDataRowLast);
    adapter.SRespRowLast(SRespRowLast);
    adapter.MByteEn(MByteEn);
    adapter.MDataByteEn(MDataByteEn);

    // TL1 interface
    adapter.ocpTL1(target.ocp);
  }

  sc_in<bool> Clk;

private:
  ocp_tl0_initiator initiator;
  ocp_tl1_target target;
  tl0_initiator_to_tl1_target<32, SRMDSEQ_TYPES> adapter;

  sc_signal<unsigned> MAddr;
  sc_signal<unsigned> MCmd;
  sc_signal<unsigned> MData;
  sc_signal<unsigned> SData;
  sc_signal<bool> SCmdAccept;
  sc_signal<unsigned> SResp;
  sc_signal<bool> MReset_n;

  sc_signal<unsigned> MBurstLength;
  sc_signal<bool> MBurstSingleReq;
  sc_signal<bool> MDataValid;
  sc_signal<bool> MDataLast;
  sc_signal<bool> SDataAccept;
  sc_signal<bool> SRespLast;
  sc_signal<unsigned> MBurstSeq;
  sc_signal<unsigned> MBlockHeight;
  sc_signal<unsigned> MBlockStride;
  sc_signal<bool> MDataRowLast;
  sc_signal<bool> SRespRowLast;
  sc_signal<unsigned> MByteEn;
  sc_signal<unsigned> MDataByteEn;
};


class testbench10: public sc_module {
public:
  testbench10(sc_module_name n):
    sc_module(n),
    Clk("Clk"),
    initiator("initiator"),
    target("target"),
    adapter("adapter")
  {
    // OCP configuration
    adapter.set_ocp_config(gen_ocp_config());

    // clock
    initiator.Clk(Clk);
    target.Clk(Clk);
    adapter.Clk(Clk);

    // TL1 interface
    initiator.ocp(adapter.ocpTL1);

    // TL0 interface
    adapter.MAddr(MAddr);
    adapter.MCmd(MCmd);
    adapter.MData(MData);
    adapter.SData(SData);
    adapter.SCmdAccept(SCmdAccept);
    adapter.SResp(SResp);
    adapter.MReset_n(MReset_n);
    target.MAddr(MAddr);
    target.MCmd(MCmd);
    target.MData(MData);
    target.SData(SData);
    target.SCmdAccept(SCmdAccept);
    target.SResp(SResp);
    target.MReset_n(MReset_n);

    adapter.MBurstLength(MBurstLength);
    adapter.MBurstSingleReq(MBurstSingleReq);
    adapter.MDataValid(MDataValid);
    adapter.MDataLast(MDataLast);
    adapter.SDataAccept(SDataAccept);
    adapter.SRespLast(SRespLast);
    adapter.MBurstSeq(MBurstSeq);
    adapter.MBlockHeight(MBlockHeight);
    adapter.MBlockStride(MBlockStride);
    adapter.MDataRowLast(MDataRowLast);
    adapter.SRespRowLast(SRespRowLast);
    adapter.MByteEn(MByteEn);
    adapter.MDataByteEn(MDataByteEn);
    target.MBurstLength(MBurstLength);
    target.MBurstSingleReq(MBurstSingleReq);
    target.MDataValid(MDataValid);
    target.MDataLast(MDataLast);
    target.SDataAccept(SDataAccept);
    target.SRespLast(SRespLast);
    target.MBurstSeq(MBurstSeq);
    target.MBlockHeight(MBlockHeight);
    target.MBlockStride(MBlockStride);
    target.MDataRowLast(MDataRowLast);
    target.SRespRowLast(SRespRowLast);
    target.MByteEn(MByteEn);
    target.MDataByteEn(MDataByteEn);
  }

  sc_in<bool> Clk;

private:
  ocp_tl1_initiator initiator;
  ocp_tl0_target target;
  tl1_initiator_to_tl0_target<32, SRMDSEQ_TYPES> adapter;

  sc_signal<unsigned> MAddr;
  sc_signal<unsigned> MCmd;
  sc_signal<unsigned> MData;
  sc_signal<unsigned> SData;
  sc_signal<bool> SCmdAccept;
  sc_signal<unsigned> SResp;
  sc_signal<bool> MReset_n;

  sc_signal<unsigned> MBurstLength;
  sc_signal<bool> MBurstSingleReq;
  sc_signal<bool> MDataValid;
  sc_signal<bool> MDataLast;
  sc_signal<bool> SDataAccept;
  sc_signal<bool> SRespLast;
  sc_signal<unsigned> MBurstSeq;
  sc_signal<unsigned> MBlockHeight;
  sc_signal<unsigned> MBlockStride;
  sc_signal<bool> MDataRowLast;
  sc_signal<bool> SRespRowLast;
  sc_signal<unsigned> MByteEn;
  sc_signal<unsigned> MDataByteEn;
};


class testbench11: public sc_module {
public:
  testbench11(sc_module_name n):
    sc_module(n),
    Clk("Clk"),
    initiator("initiator"),
    target("target")
  {
    // clock
    initiator.Clk(Clk);
    target.Clk(Clk);

    // TL1 interface
    initiator.ocp(target.ocp);
  }

  sc_in<bool> Clk;

private:
  ocp_tl1_initiator initiator;
  ocp_tl1_target target;
};


class testbench101: public sc_module {
public:
  testbench101(sc_module_name n):
    sc_module(n),
    Clk("Clk"),
    initiator("initiator"),
    target("target"),
    t_print(&cout),
    adapter10("adapter10", &t_print),
    adapter01("adapter01", &t_print)
  {
    // OCP configuration
    adapter10.set_ocp_config(gen_ocp_config());
    adapter01.set_ocp_config(gen_ocp_config());

    // clock
    initiator.Clk(Clk);
    target.Clk(Clk);
    adapter10.Clk(Clk);
    adapter01.Clk(Clk);

    // TL1 interface
    initiator.ocp(adapter10.ocpTL1);
    adapter01.ocpTL1(target.ocp);

    // TL0 interface
    adapter10.MAddr(MAddr);
    adapter10.MCmd(MCmd);
    adapter10.MData(MData);
    adapter10.SData(SData);
    adapter10.SCmdAccept(SCmdAccept);
    adapter10.SResp(SResp);
    adapter10.MReset_n(MReset_n);
    adapter01.MAddr(MAddr);
    adapter01.MCmd(MCmd);
    adapter01.MData(MData);
    adapter01.SData(SData);
    adapter01.SCmdAccept(SCmdAccept);
    adapter01.SResp(SResp);
    adapter01.MReset_n(MReset_n);

    adapter10.MBurstLength(MBurstLength);
    adapter10.MBurstSingleReq(MBurstSingleReq);
    adapter10.MDataValid(MDataValid);
    adapter10.MDataLast(MDataLast);
    adapter10.SDataAccept(SDataAccept);
    adapter10.SRespLast(SRespLast);
    adapter10.MBurstSeq(MBurstSeq);
    adapter10.MBlockHeight(MBlockHeight);
    adapter10.MBlockStride(MBlockStride);
    adapter10.MDataRowLast(MDataRowLast);
    adapter10.SRespRowLast(SRespRowLast);
    adapter10.MByteEn(MByteEn);
    adapter10.MDataByteEn(MDataByteEn);
    adapter01.MBurstLength(MBurstLength);
    adapter01.MBurstSingleReq(MBurstSingleReq);
    adapter01.MDataValid(MDataValid);
    adapter01.MDataLast(MDataLast);
    adapter01.SDataAccept(SDataAccept);
    adapter01.SRespLast(SRespLast);
    adapter01.MBurstSeq(MBurstSeq);
    adapter01.MBlockHeight(MBlockHeight);
    adapter01.MBlockStride(MBlockStride);
    adapter01.MDataRowLast(MDataRowLast);
    adapter01.SRespRowLast(SRespRowLast);
    adapter01.MByteEn(MByteEn);
    adapter01.MDataByteEn(MDataByteEn);

    // timing:  TL1 initiator and target produce transport calls at +0 ps
    //   but they aren't seen by adapters until +1 ps because of synch protection
    //   and only at this time are TL0 signals written
    adapter01.set_sample_times(
      sc_time(2.0, SC_PS),
      // adapter10 writes MCmd always during TLM-transport-from-initiator at +1 ps
      sc_time(3.0, SC_PS)
      // same for MDataValid but sampling is delayed to be after MCmd sampling,
      // which would be automatic but shown here for clarity
    );
    adapter10.set_sample_times(
      sc_time(2.0, SC_PS),
      // adapter01 writes SResp always during TLM-transport-from-target at +1 ps
      sc_time(5.0, SC_PS),
      // TL1 target may signal SCmdAccept during its clocked process with an
      // explicit transport call.  This would be seen at +1 ps by adapter01 and
      // the TL0 SCmdAccept signal would be stable for sampling at +2 ps.  But
      // TL1 target might return SCmdAccept to a fw transport call, which it
      // would see at the MCmd sampling time of adapter01 plus 1, ie 3 ps.  The
      // socket PEQ will convert this into a separate backward transport call
      // seen by adapter01 at 4 ps.  Thus SCmdAccept should be sampled at 5 ps.
      sc_time(6.0, SC_PS)
      // data trails command by 1 ps
    );
  }

  sc_in<bool> Clk;

private:
  ocp_tl1_initiator initiator;
  ocp_tl1_target target;

  tl1_tl0::get_TL0_timing_example<ostream> t_print;
  tl1_initiator_to_tl0_target<32, SRMDSEQ_TYPES> adapter10;
  tl0_initiator_to_tl1_target<32, SRMDSEQ_TYPES> adapter01;

  sc_signal<unsigned> MAddr;
  sc_signal<unsigned> MCmd;
  sc_signal<unsigned> MData;
  sc_signal<unsigned> SData;
  sc_signal<bool> SCmdAccept;
  sc_signal<unsigned> SResp;
  sc_signal<bool> MReset_n;

  sc_signal<unsigned> MBurstLength;
  sc_signal<bool> MBurstSingleReq;
  sc_signal<bool> MDataValid;
  sc_signal<bool> MDataLast;
  sc_signal<bool> SDataAccept;
  sc_signal<bool> SRespLast;
  sc_signal<unsigned> MBurstSeq;
  sc_signal<unsigned> MBlockHeight;
  sc_signal<unsigned> MBlockStride;
  sc_signal<bool> MDataRowLast;
  sc_signal<bool> SRespRowLast;
  sc_signal<unsigned> MByteEn;
  sc_signal<unsigned> MDataByteEn;
};


class testbench010: public sc_module {
public:
  testbench010(sc_module_name n):
    sc_module(n),
    Clk("Clk"),
    initiator("initiator"),
    target("target"),
    adapter01("adapter01"),
    adapter10("adapter10")
  {
    // OCP configuration
    adapter01.set_ocp_config(gen_ocp_config());
    adapter10.set_ocp_config(gen_ocp_config());

    // clock
    initiator.Clk(Clk);
    target.Clk(Clk);
    adapter01.Clk(Clk);
    adapter10.Clk(Clk);

    // TL1 interface
    adapter01.ocpTL1(adapter10.ocpTL1);

    // TL0 interfaces
    initiator.MAddr(MAddr[0]);
    initiator.MCmd(MCmd[0]);
    initiator.MData(MData[0]);
    initiator.SData(SData[0]);
    initiator.SCmdAccept(SCmdAccept[0]);
    initiator.SResp(SResp[0]);
    initiator.MReset_n(MReset_n[0]);
    adapter01.MAddr(MAddr[0]);
    adapter01.MCmd(MCmd[0]);
    adapter01.MData(MData[0]);
    adapter01.SData(SData[0]);
    adapter01.SCmdAccept(SCmdAccept[0]);
    adapter01.SResp(SResp[0]);
    adapter01.MReset_n(MReset_n[0]);

    initiator.MBurstLength(MBurstLength[0]);
    initiator.MBurstSingleReq(MBurstSingleReq[0]);
    initiator.MDataValid(MDataValid[0]);
    initiator.MDataLast(MDataLast[0]);
    initiator.SDataAccept(SDataAccept[0]);
    initiator.SRespLast(SRespLast[0]);
    initiator.MBurstSeq(MBurstSeq[0]);
    initiator.MBlockHeight(MBlockHeight[0]);
    initiator.MBlockStride(MBlockStride[0]);
    initiator.MDataRowLast(MDataRowLast[0]);
    initiator.SRespRowLast(SRespRowLast[0]);
    initiator.MByteEn(MByteEn[0]);
    initiator.MDataByteEn(MDataByteEn[0]);
    adapter01.MBurstLength(MBurstLength[0]);
    adapter01.MBurstSingleReq(MBurstSingleReq[0]);
    adapter01.MDataValid(MDataValid[0]);
    adapter01.MDataLast(MDataLast[0]);
    adapter01.SDataAccept(SDataAccept[0]);
    adapter01.SRespLast(SRespLast[0]);
    adapter01.MBurstSeq(MBurstSeq[0]);
    adapter01.MBlockHeight(MBlockHeight[0]);
    adapter01.MBlockStride(MBlockStride[0]);
    adapter01.MDataRowLast(MDataRowLast[0]);
    adapter01.SRespRowLast(SRespRowLast[0]);
    adapter01.MByteEn(MByteEn[0]);
    adapter01.MDataByteEn(MDataByteEn[0]);

    adapter10.MAddr(MAddr[1]);
    adapter10.MCmd(MCmd[1]);
    adapter10.MData(MData[1]);
    adapter10.SData(SData[1]);
    adapter10.SCmdAccept(SCmdAccept[1]);
    adapter10.SResp(SResp[1]);
    adapter10.MReset_n(MReset_n[1]);
    target.MAddr(MAddr[1]);
    target.MCmd(MCmd[1]);
    target.MData(MData[1]);
    target.SData(SData[1]);
    target.SCmdAccept(SCmdAccept[1]);
    target.SResp(SResp[1]);
    target.MReset_n(MReset_n[1]);

    adapter10.MBurstLength(MBurstLength[1]);
    adapter10.MBurstSingleReq(MBurstSingleReq[1]);
    adapter10.MDataValid(MDataValid[1]);
    adapter10.MDataLast(MDataLast[1]);
    adapter10.SDataAccept(SDataAccept[1]);
    adapter10.SRespLast(SRespLast[1]);
    adapter10.MBurstSeq(MBurstSeq[1]);
    adapter10.MBlockHeight(MBlockHeight[1]);
    adapter10.MBlockStride(MBlockStride[1]);
    adapter10.MDataRowLast(MDataRowLast[1]);
    adapter10.SRespRowLast(SRespRowLast[1]);
    adapter10.MByteEn(MByteEn[1]);
    adapter10.MDataByteEn(MDataByteEn[1]);
    target.MBurstLength(MBurstLength[1]);
    target.MBurstSingleReq(MBurstSingleReq[1]);
    target.MDataValid(MDataValid[1]);
    target.MDataLast(MDataLast[1]);
    target.SDataAccept(SDataAccept[1]);
    target.SRespLast(SRespLast[1]);
    target.MBurstSeq(MBurstSeq[1]);
    target.MBlockHeight(MBlockHeight[1]);
    target.MBlockStride(MBlockStride[1]);
    target.MDataRowLast(MDataRowLast[1]);
    target.SRespRowLast(SRespRowLast[1]);
    target.MByteEn(MByteEn[1]);
    target.MDataByteEn(MDataByteEn[1]);

    // timing: nothing needed as adapters' TL0 interfaces are
    //  directly connected to TL0-native initiator and target
  }

  sc_in<bool> Clk;

private:
  ocp_tl0_initiator initiator;
  ocp_tl0_target target;
  tl0_initiator_to_tl1_target<32, SRMDSEQ_TYPES> adapter01;
  tl1_initiator_to_tl0_target<32, SRMDSEQ_TYPES> adapter10;

  sc_signal<unsigned> MAddr[2];
  sc_signal<unsigned> MCmd[2];
  sc_signal<unsigned> MData[2];
  sc_signal<unsigned> SData[2];
  sc_signal<bool> SCmdAccept[2];
  sc_signal<unsigned> SResp[2];
  sc_signal<bool> MReset_n[2];

  sc_signal<unsigned> MBurstLength[2];
  sc_signal<bool> MBurstSingleReq[2];
  sc_signal<bool> MDataValid[2];
  sc_signal<bool> MDataLast[2];
  sc_signal<bool> SDataAccept[2];
  sc_signal<bool> SRespLast[2];
  sc_signal<unsigned> MBurstSeq[2];
  sc_signal<unsigned> MBlockHeight[2];
  sc_signal<unsigned> MBlockStride[2];
  sc_signal<bool> MDataRowLast[2];
  sc_signal<bool> SRespRowLast[2];
  sc_signal<unsigned> MByteEn[2];
  sc_signal<unsigned> MDataByteEn[2];
};


class testbench: public sc_module {
public:
  testbench(sc_module_name n, bool run_01 = true, bool run_10 = true):
    sc_module(n),
    tb00("00"),
    tb01("01"),
    tb10("10"),
    tb11("11"),
    tb101("101"),
    tb010("010"),
    Clk("Clk", sc_time(10.0, SC_NS))
  {
    tb00.Clk(Clk);
    tb11.Clk(Clk);

    if(run_01) tb01.Clk(Clk);
    else tb01.Clk(tieoff);

    if(run_10) tb10.Clk(Clk);
    else tb10.Clk(tieoff);

    if(run_01 && run_10) {
      tb010.Clk(Clk);
      tb101.Clk(Clk);
    } else {
      tb101.Clk(tieoff);
      tb010.Clk(tieoff);
    }
  }
private:
  testbench00 tb00;
  testbench01 tb01;
  testbench10 tb10;
  testbench11 tb11;
  testbench101 tb101;
  testbench010 tb010;
  sc_clock Clk;
  sc_signal<bool> tieoff;
};


#ifndef OMIT_SC_MAIN
int sc_main(int argc, char **argv) {
  sc_report_handler::set_actions(SC_ERROR, SC_DISPLAY | SC_ABORT);
  bool run_01 = ((argc < 2) || ('A' == *argv[1]) || ('0' == *argv[1]));
  bool run_10 = ((argc < 2) || ('A' == *argv[1]) || ('1' == *argv[1]));
  testbench theTestbench("testbench", run_01, run_10);
  sc_start();
  return 0;
}
#endif

