// 
//  Copyright 2003 OCP-IP
//
// ============================================================================
//      Project : OCP SLD WG, OCP Transaction Level Layer-1
//       Author : Norman Weyrich, Synopsys Inc., weyrich@synopsys.com
//                Anssi Haverinen, Nokia Inc., anssi.haverinen@nokia.com
//                Yann Bajot, Prosilog, bajot@prosilog.com
//         Date : 01/15/2004
//
//  Description : Transaction Level - Layer-1 example Master (OCP 2.0)
//               (with Datahandshake support)
//
//    Features:
//    - Asynchronous master using OCP 2.0 API
//    - Synchronous Datahandshake , with propagation delay emulation 
//    - Non-blocking methods are used
//    - Pipelined Request and Response threads
//    - Performs 5 WRITE operations followed by 5 READ operations
//       and checks read data, assuming that the target
//       acts as a RAM
//                
// Parameter    :
//   Template arguments.
//   - TdataCl: Data class containing data members and
//              access methods for the data exchanged between
//              Masters and Slaves
//   - Td: Data type
//   - Ta: Address type
//                
//   Constructor arguments.
//     - sc_module_name: Name of the module instance
//     - ID: Unique number identifying the Master.
//           ID must be unique among all Masters attached to the same Bus.
//     - Priority: Positive number specifying the priority relative to the
//                 other Masters for Bus access. Higher numbers means higher
//                 priority. Masters can have the same priority.
//     - RequestDelay: Non-negative number specifying the delay in percent
//                 of a clock cycle, the request signals are transmitted.
//     - DataDelay: Non-negative number specifying the delay in percent
//                 of a clock cycle, the Datahandshake signals are transmitted
//                 after the clock rising edge following the request.
//     - ResponseDelay: Non-negative number specifying the delay in percent
//                 of a clock cycle, the response is acknowledged .
// ============================================================================

// Used for debug purposes
#define DEBUG_TL1_MASTER_ASYNC_HS

#include "ocp_tl1_master_async_hs.h"
#include "ocp_tl1_data_cl.h"

// ----------------------------------------------------------------------------
// Process : OCP_TL1_Master_Async_hs::OCP_TL1_Master_Async_hs
// 
// Constructor
// ----------------------------------------------------------------------------
template<class TdataCl> OCP_TL1_Master_Async_hs<TdataCl>::OCP_TL1_Master_Async_hs
      (
        sc_module_name name_
      , int ID
      , int Priority
      , double RequestDelay
      , double DataDelay
      , double ResponseDelay
      )
        : sc_module         (name_)
        , MasterP           ("MasterPort")
        , m_ID              (ID)
        , m_Priority        (Priority)
        , m_RequestDelay    (RequestDelay)
        , m_DataDelay       (DataDelay)
        , m_ResponseDelay   (ResponseDelay)
{
  // InitParameters();

  SC_THREAD(MasterRequest);
  sensitive_pos(clk); 
  SC_THREAD(MasterResponse);
  sensitive << MasterP; 
}

// ----------------------------------------------------------------------------
// Process : OCP_TL1_Master_Async_hs::~OCP_TL1_Master_Async_hs
// 
// Destructor
// ----------------------------------------------------------------------------
template<class TdataCl> OCP_TL1_Master_Async_hs<TdataCl>::~OCP_TL1_Master_Async_hs()
{
}


// ----------------------------------------------------------------------------
//  Method : OCP_TL1_Master_Async_hs::MasterRequest()
//
//  Top level request method of the Master
//  - Issues Request calls, and Datahandshake calls one cycle later
// ----------------------------------------------------------------------------
template<class TdataCl> void OCP_TL1_Master_Async_hs<TdataCl>::MasterRequest()
{
  int i = 0;
  int j;
  int Pause[] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0}; // # of cycles between requests;
  //int Pause[] = {1, 1, 1, 1, 1, 1, 1, 1, 1, 1}; // # of cycles between requests;
  //int Pause[] = {2, 2, 2, 2, 2, 2, 2, 2, 2, 2}; // # of cycles between requests;

  Ta Addr[] =     {0,4,8,12,16,0,4,8,12,16};    // MAddr field for each request
  Td SentData[] =     {5,4,3,2,1,0,0,0,0,0};    // MData field for each request
  int cnt = 0;

  // Reset the TL1 request and dataHS structures
  TL1_request.reset();
  TL1_dataHS.reset();
  

  while(true)
  {
    wait(); // wait for rising edge of clock

    if (cnt == 0)
    {
      // check if channel is free
      if (!MasterP->getSBusy())
      {
        // get new Pause value
        j=i;
        i++;
        cnt = Pause[i%10];

        // Delay request signals
        // This must work with or without the next line
        if (m_RequestDelay > 0)
          wait(TL_CLK_PERIOD * (m_RequestDelay/100.0), TL_TIME_UNIT);

        // Put request in the channel
        TL1_request.MAddr = Addr[j%10];
        if(j%10<5) {
          TL1_request.MCmd = OCP_MCMD_WR; // WRITE operations 
        }
        else {
          TL1_request.MCmd = OCP_MCMD_RD; // READ operations 
        }
        
        // Sent the TL1 request to the channel
        bool ret = MasterP->startOCPRequest(TL1_request);

        // For Debug purposes only
        assert(ret);

#ifdef DEBUG_TL1_MASTER_ASYNC_HS
        cout << "TL1 Master sent request "
          << " delta " << simcontext()->delta_count()
          << " time " << sc_time_stamp().to_seconds()
          << " addr " << TL1_request.MAddr;
        cout << endl;
#endif

        // Put data in the channel (Datahandshake)
        if(j%10<5) {
          do {
            wait(clk.posedge_event());
          } while(MasterP->getSBusyDataHS());

          // Delay Datahandshake signals
          // This must work with or without the next line
          if (m_DataDelay > 0) {
            wait(TL_CLK_PERIOD * (m_DataDelay/100.0), TL_TIME_UNIT);
          }
          TL1_dataHS.MData = SentData[j%10];
          TL1_dataHS.MDataValid = true; // Not mandatory

          MasterP->startOCPDataHS(TL1_dataHS); // Toggles the handshake data members

#ifdef DEBUG_TL1_MASTER_ASYNC_HS
          cout << "TL1 Master sent data "
            << " delta " << simcontext()->delta_count()
            << " time " << sc_time_stamp().to_seconds()
            << " written data " << SentData[j%10] << endl;
#endif
        }

      }
      else
      {
        cnt--;
      }
    }
  } // end of while loop
}


// ----------------------------------------------------------------------------
//  Method : OCP_TL1_Master_Async_hs::MasterResponse()
//
//  Top level response method of the Master
//
// ----------------------------------------------------------------------------
template<class TdataCl> void OCP_TL1_Master_Async_hs<TdataCl>::MasterResponse()
{
#ifdef DEBUG_TL1_MASTER_ASYNC_HS
  Td CorrectData[] =  {5,4,3,2,1};
#endif

  int i = 0;

  while(true)
  {
    wait(); // for default event

    if (MasterP->getOCPResponse(TL1_response,false))
    {
    // There is a valid response on the channel

#ifdef DEBUG_TL1_MASTER_ASYNC_HS
      cout << "TL1 Master got response "
           << " delta " << simcontext()->delta_count()
           << " time " << sc_time_stamp().to_seconds()
           << " received data " << TL1_response.SData ;
      if(TL1_response.SData != CorrectData[i%5]) 
        cout << " is BAD" << endl;
      else
        cout << " is good" << endl;
#endif

      // Models the response acknowledge delay
      if (m_ResponseDelay > 0)
        wait(TL_CLK_PERIOD * (m_ResponseDelay/100.0), TL_TIME_UNIT);
        
      // Release the response 
      MasterP->putMRespAccept();

#ifdef DEBUG_TL1_MASTER_ASYNC_HS
      cout << "TL1 Master released response "
           << " delta " << simcontext()->delta_count()
           << " time " << sc_time_stamp().to_seconds()
           << " data " << TL1_response.SData << endl;
#endif

      i++;
    } else {
      // No response available 
    }
  } // end of while(true) loop
}


// ----------------------------------------------------------------------------
//  Method : OCP_TL1_Master_Async_hs::SputDirect()
//
//  Debug interface method.
//  Read/Write data, Slave to Master direction
//  Returns true for success and false for failure
//
// ----------------------------------------------------------------------------
template<class TdataCl> bool OCP_TL1_Master_Async_hs<TdataCl>::SputDirect(
         int SlaveID, bool IsWrite, Td *Data, Ta Address, int NumWords)

{
  // not implemented
  return(false);
}


// ----------------------------------------------------------------------------
//  Method : OCP_TL1_Master_Async_hs::end_of_elaboration()
//
//  This method is activated at the end of the elaboration phase
//  (when all binding operations are finished).
// ----------------------------------------------------------------------------
template<class TdataCl> void OCP_TL1_Master_Async_hs<TdataCl>::end_of_elaboration()
{
  sc_module::end_of_elaboration();

  // Initialize debug interface
  MasterP->MregisterDirectIF(this);

  // Put parameter into Channel
  // No parameters needed
}


// ----------------------------------------------------------------------------
//
//  Instantiation of the Master
//
// ----------------------------------------------------------------------------
template class OCP_TL1_Master_Async_hs<OCP_TL1_DataCl<int,int> >;
