// 
//  Copyright 2003 OCP-IP
//  OCP-IP Confidential & Proprietary
//
// ============================================================================
//      Project : OCP SLD WG, Layer adapter examples 
//      Authors : Alan Kamas based on previous work and data traffic pattern by Yann Bajot
//          $Id:
//
//  Description:  Simple Point to point TL2 example using the new OCP-specific
//                TL2 API using static events only - no threads.
// ============================================================================

//#define DEBUG_C
//#define TRACE_C

#include <systemc.h>
#include "ocp_tl2_channel.h"  
#include "ocp_tl2_master_port.h"  
#include "ocp_tl2_slave_port.h"  

// ----------------------------------------------------------------------------
// SIMPLE MASTER MODULE:
//
//  * non-pipelined module: no threads - events only. 
//  * uses an OCP-specific TL2 master port
//  * uses the new 'OCPTL2RequestGrp' structure to pass all the request 
//      signals to the channel, and a 'OCPTL2ResponseGrp' structure to 
//      store the response signals
//  * uses methods only: no threads
//  * uses three SC_Methods and one internal event.
//  * performs the following actions in a loop:
//      - sends a 10-length WRITE burst to the slave using
//        sendOCPRequest(). 
//      - sends a 10-length READ burst to the slave using
//        sendOCPRequest(). 
//      - gets the corresponding responses.
//      - performs a 20-length WRITE request.
//      - performs a 20-length READ request and gets the corresponding response.
// 
// ----------------------------------------------------------------------------

// const int LOOP_COUNT = 5;
const int LOOP_COUNT = 10000000;
// const int LOOP_COUNT = 1000000;
const int END_TIME = 1000000000;

SC_MODULE(master) {

    // this module has SC processes
    SC_HAS_PROCESS(master);

    // OCP-specific TL2 specialized master port 
    OCP_TL2_MasterPort<unsigned int, unsigned int> ocp;

    // ----------------------------------------------------------------------------
    // Request Method
    // ----------------------------------------------------------------------------

    // This method is called when the ocp connection is clear for a new request
    void send_request_method () 
    {
        // cout << "Master starts sending request " << m_req_num << " -- at time " << sc_time_stamp() << endl;
        switch(m_req_num) 
        {
            case 0:
                cout << "master DONE. \n" << endl;
                return;

            case 1:
                // Send the first reqeust
                m_req.reset();
                // Set the members of the request group stucture
                m_req.MCmd            =     OCP_MCMD_WR;    // this is a WRITE transaction
                m_req.MDataPtr        =     wdata;          // pass a pointer to the master data array
                // Sets TL2-specific fields
                // this request is sent using a single TL2 chunk
                // Original TL2 API: ReqChunkLen = 10; ReqChunkLast = true;
                m_req.DataLength = 10;
                m_req.LastOfBurst = true;
                // set request to send next time
                m_req_num = 2;
                break;

            case 2:
                // Send the second request
                // Set the members of the request group stucture that have changed
                m_req.MCmd            =     OCP_MCMD_RD;    // this is a READ transaction
                // Original TL2 API: ReqChunkLen = 10; ReqChunkLast = true;
                m_req.DataLength = 10;
                m_req.LastOfBurst = true;

                // set request to send next time
                m_req_num=3;
                break;

            case 3:
                // Send the third request
                // Set the members of the request group stucture that have changed
                m_req.MCmd            =     OCP_MCMD_WR;    // this is a READ transaction
                m_req.MDataPtr        =     wdata;          // pass a pointer to the master data array
                // Original TL2 API: ReqChunkLen = 20; ReqChunkLast = true;
                m_req.DataLength = 20;
                m_req.LastOfBurst = true;

                // set request to send next time
                m_req_num=4;
                break;

            case 4:
                // Send the fourth request
                // Set the members of the request group stucture that have changed
                m_req.MCmd            =     OCP_MCMD_RD;    // this is a READ transaction
                // Original TL2 API: ReqChunkLen = 20; ReqChunkLast = true;
                m_req.DataLength = 20;
                m_req.LastOfBurst = true;

                // set request to send next time
                m_loops_done++;
                if (m_loops_done >= LOOP_COUNT)
                {
                    m_req_num = 0;
                } 
                else 
                {
                    m_req_num = 1;
                }
                break;

            default:
                cout << "ERROR: master reached unknown request number = " << m_req_num << endl;
        }
        // Now that the request has been created, send it.

        // Original TL2 API:
        // bool ret_value=ocp->sendOCPRequest(m_req,ReqChunkLen,ReqChunkLast);

        // New TL2 API:
        bool ret_value=ocp->sendOCPRequest(m_req);

        assert(ret_value);
        // cout << "Master ends sending request at time " << sc_time_stamp() << endl;
    }
        
    // when this method is called there should be a waiting response on the ocp connection
    void get_response_method () 
    {
        switch( m_resp_num )
        {
            case 1:
                // cout << "Master starts getting response 1, chunk 1 -- " << sc_time_stamp() << endl;
                
                // Original TL2 API:
                // release_immediately = false;
                // ret_value=ocp->getOCPResponse(m_resp,release_immediately,RespChunkLen,RespChunkLast);
                // assert(RespChunkLen == 5); assert(RespChunkLast == false);

                // New TL2 API:
                ret_value=ocp->getOCPResponse(m_resp);
                assert(ret_value);
                // This a small regression test. 
                assert(m_resp.DataLength == 5); 
                assert(m_resp.LastOfBurst == false);

                // Display the received data
                // for ( unsigned int i=0; i< RespChunkLen; i+=1 ) {
                    // cout << m_resp.SDataPtr[i] << " " ;
                // }
                // cout << endl;
                // release the response after some delay
                m_releaseResponseEvent.notify(100,SC_NS);
                // set up the next response
                m_resp_num =2;
                break;

            case 2:
                // cout << "Master starts getting response 1, chunk 2 -- " << sc_time_stamp() << endl;
                // Original OCP TL2 API:
                // release_immediately = false;
                // ret_value=ocp->getOCPResponse(m_resp,release_immediately,RespChunkLen,RespChunkLast);
                // assert(RespChunkLen == 5); assert(RespChunkLast == true);

                // New OCP TL2 API:
                ret_value=ocp->getOCPResponse(m_resp);
                assert(m_resp.DataLength == 5); 
                assert(m_resp.LastOfBurst == true);
                assert(ret_value);

                // Display the received data
                // for ( unsigned int i=0; i< RespChunkLen; i+=1 ) {
                    // cout << m_resp.SDataPtr[i] << " " ;
                // }
                // cout << endl;
                // release the response after some delay
                m_releaseResponseEvent.notify(100,SC_NS);
                // set up the next response
                m_resp_num =3;
                break;

            case 3:
                // cout << "Master starts getting response 2  -- " << sc_time_stamp() << endl;
                // Original OCP TL2 API:
                // release_immediately = true;
                // ret_value=ocp->getOCPResponse(m_resp,release_immediately,RespChunkLen,RespChunkLast);
                // assert(RespChunkLen == 20); assert(RespChunkLast == true);

                // New OCP TL2 API:
                ret_value=ocp->getOCPResponse(m_resp);
                assert(m_resp.DataLength == 20); 
                assert(m_resp.LastOfBurst == true);
                assert(ret_value);
                // Accept immediately (not accurate OCP timing)
                ocp->acceptResponse();

                // Display the received data
                // for ( unsigned int i=0; i< RespChunkLen; i+=1 ) {
                    // cout << m_resp.SDataPtr[i] << " " ;
                // }
                // cout << endl;
                // response should be released automatically.
                // set up the next response
                // set response to send next time
                m_resp_num = 1;
                break;

            default:
                cout << "ERROR: master reached unknown response number = " << m_resp_num << endl;
        }

    }

    // when this method is called there should be a waiting response on the ocp connection
    void release_response_method () 
    {
        // Original OCP TL2 API:
        // ocp->putMRespAccept();

        // New OCP TL2 API:
        ocp->acceptResponse();

    }

    // ----------------------------------------------------------------------------
    // Constuctor
    // ----------------------------------------------------------------------------
    master(sc_module_name mod):
        sc_module(mod), ocp("Master_Port")
    {
        // declare my three sensitive methods
        SC_METHOD(send_request_method);
          sensitive << ocp.RequestEndEvent();
        SC_METHOD(get_response_method); 
          sensitive << ocp.ResponseStartEvent();
          dont_initialize();
        SC_METHOD(release_response_method); 
          sensitive << m_releaseResponseEvent;
          dont_initialize();

        // Init Master Data array
        for (unsigned int i=0;i<100;i++)
        {
            wdata[i] = i;
        }
        m_req_num = 1;   // First request to send
        m_loops_done = 0;
        m_resp_num = 1;   // First response to get
        // ReqChunkLen = 10;  // No longer needed
        // ReqChunkLast = true;  // No longer needed
        ret_value = false;
        // release_immediately = false;  // No longer needed
        // RespChunkLen = 10;  // No longer needed
        // RespChunkLast = true; // No longer needed
    }

    // ----------------------------------------------------------------------------
    // Internal class members 
    // ----------------------------------------------------------------------------
    
    public:

    private:

    // Array storing data to be sent
    unsigned int wdata[100];

    // Release event for delayed releases
    sc_event m_releaseResponseEvent;

    // OCP Request/Response structures
    int m_req_num;
    int m_loops_done;
    // unsigned int ReqChunkLen; // No longer needed
    // bool ReqChunkLast;// No longer needed
    bool ret_value;
    int m_resp_num;
    // bool release_immediately;// No longer needed
    // unsigned int RespChunkLen;// No longer needed
    // bool RespChunkLast;// No longer needed
    OCPTL2RequestGrp<unsigned int, unsigned int> m_req;
    OCPTL2ResponseGrp<unsigned int> m_resp;

};

// ----------------------------------------------------------------------------
// SIMPLE SLAVE MODULE:
// 
//  * non-pipelined module: event drven only - no SC_Thread
//  * uses an OCP-specific TL2 slave port
//  * uses an'OCPTL2RequestGrp' structure to store all the request bursts
//      and an 'OCPTL2ResponseGrp' structure to send response bursts
//  * performs the following actions:
//      - receives a 10-length WRITE burst from the master, and stores the
//          received data in an internal array.
//      - receives a 10-length READ burst from the master, and sends the
//        response using two consecutive response chunks (5-length each) with a
//        different 'SRespInfo' signal value.
//      - receives a 20-length WRITE burst from the master, and stores the
//          received data in an internal array.
//      - receives a 20-length READ burst from the master,and sends the
//        response using one response call
// 
// ----------------------------------------------------------------------------

SC_MODULE(slave) {

    // OCP-specific TL2 specialized master port 
    OCP_TL2_SlavePort<unsigned int,unsigned int> ocp;

    // this module has SC processes
    SC_HAS_PROCESS(slave);

    // Get request and generate response
    // when this method is called there should be a waiting request on the ocp connection
    void get_request_method () 
    {
        switch( m_req_num )
        {
            case 1:
                // ---------------------------------------------------------
                // Gets the first request (should be a 10-length WRITE)
                // ---------------------------------------------------------
                //  We will release the request channel after some delay
                // cout << "Slave starts getting request 1 -- " << sc_time_stamp() << endl;

                // Original OCP TL2 API:
                // get the first request chunk and retrieves chunk parameters
                // release_immediately = false;
                // ret_value=ocp->getOCPRequest(m_req,release_immediately,ReqChunkLen,ReqChunkLast);
                // assert(ReqChunkLen == 10); assert(ReqChunkLast == true);

                // New OCP TL2 API:
                ret_value=ocp->getOCPRequest(m_req);
                assert(m_req.DataLength == 10); 
                assert(m_req.LastOfBurst == true);

                assert(ret_value); 
                assert(m_req.MCmd == OCP_MCMD_WR);

                // Copy the received data into the internal buffer
                for (unsigned int i=0; i< m_req.DataLength; i+=1 ) {
                    data[i] = m_req.MDataPtr[i];
                }

                // releases the request after some delay 
                m_releaseRequestEvent.notify(100,SC_NS);

                // set up the next request
                m_req_num =2;
                break;

            case 2:
                // ---------------------------------------------------------
                // Gets the second request (10-length READ) 
                // ---------------------------------------------------------
                // Now we release the request immediately
                // cout << "Slave starts getting request 2 -- " << sc_time_stamp() << endl;

                // Original OCP TL2 API:
                // get the first request chunk and retrieves the chunk parameters
                // release_immediately = true;
                // ret_value=ocp->getOCPRequestBlocking(m_req,release_immediately,ReqChunkLen,ReqChunkLast);
                // assert(ReqChunkLen == 10); assert(ReqChunkLast == true);

                // New OCP TL2 API:
                ret_value=ocp->getOCPRequestBlocking(m_req);
                assert(m_req.DataLength == 10); 
                assert(m_req.LastOfBurst == true);

                // This a small regression test. 
                assert(ret_value); assert(m_req.MCmd == OCP_MCMD_RD);

                // Accept Immediately (even though this violates OCP timing)
                ocp->acceptRequest();

                // cout << "Slave ends getting request 2 -- " << sc_time_stamp() << endl;

                // ---------------------------------------------------------
                // Sends the response using 2 chunks (Resp 1 & Resp 2)
                // ---------------------------------------------------------

                // Setup the first chunk 
                // Set the members of the response group stucture
                m_resp[1].SResp          =     OCP_SRESP_DVA;      // Response is VALID
                m_resp[1].SDataPtr       =     data;               // pass a pointer to the slave data
                m_resp[1].SRespInfo      =     1;                  // Info for the first chunk
                // Original OCP TL2 API:
                // assert(RespChunkLen[1] == 0);
                // RespChunkLen[1] = 5; RespChunkLast[1] = false;

                // New OCP TL2 API:
                assert(m_resp[1].DataLength == 0);
                m_resp[1].DataLength = 5;
                m_resp[1].LastOfBurst = false;

                // Setup the second chunk 
                m_resp[2].SDataPtr       =     &data[5];           // pass a pointer to the second part of the slave data
                m_resp[2].SRespInfo      =     2;                  // Info for the second chunk

                // this is the last chunk of the response burst

                // Original OCP TL2 API:
                // assert(RespChunkLen[2] == 0);
                // RespChunkLen[2] = 5; RespChunkLast[2] = true;         

                // New OCP TL2 API:
                assert(m_resp[2].DataLength == 0);
                m_resp[2].DataLength = 5;
                m_resp[2].LastOfBurst = true;

                // try to ship these  new respones out
                tryToSendResponse();
                // set up the next request
                m_req_num =3;
                break;

            case 3:

                // ---------------------------------------------------------
                // gets the third request (20-length WRITE)
                // ---------------------------------------------------------
                //  We will release the request channel after some delay


                // cout << "Slave starts getting request 3 -- " << sc_time_stamp() << endl;

                // Original OCP TL2 API:
                // get the first request chunk and retrieves chunk parameters
                // release_immediately = false;
                // ret_value=ocp->getOCPRequest(m_req,release_immediately,ReqChunkLen,ReqChunkLast);
                // assert(ReqChunkLen == 20); assert(ReqChunkLast == true);

                // New OCP TL2 API:
                ret_value=ocp->getOCPRequest(m_req);
                assert(m_req.DataLength == 20); 
                assert(m_req.LastOfBurst == true);

                // This a small regression test. 
                assert(ret_value); 
                assert(m_req.MCmd == OCP_MCMD_WR);

                // Copy the received data into the internal buffer
                for (unsigned int i=0; i< m_req.DataLength; i+=1 ) {
                    data[i] = m_req.MDataPtr[i];
                }

                // releases the request after some delay 
                m_releaseRequestEvent.notify(100,SC_NS);

                // cout << "Slave ends getting request 3 -- " << sc_time_stamp() << endl;
                // set up the next request
                m_req_num = 4;
                break;

            case 4:
                // ---------------------------------------------------------
                // gets the fourth request (20-length READ) 
                // ---------------------------------------------------------
                
                // Now we release the request immediately
    
                // gets the first request chunk and retrieves chunk parameters
                // cout << "Slave starts getting request 4 -- " << sc_time_stamp() << endl;

                // Original OCP TL2 API:
                // release_immediately = true;
                // ret_value=ocp->getOCPRequest(m_req,release_immediately,ReqChunkLen,ReqChunkLast);
                // assert(ReqChunkLen == 20); assert(ReqChunkLast == true);

                // New OCP TL2 API:
                ret_value=ocp->getOCPRequest(m_req);
                assert(m_req.DataLength == 20); 
                assert(m_req.LastOfBurst == true);
                assert(ret_value);
                // Accept immediately (even though this violates OCP timing)
                ocp->acceptRequest();

                // Regression test
                assert(m_req.MCmd == OCP_MCMD_RD);

                // cout << "Slave ends getting request 4 -- " << sc_time_stamp() << endl;
                // Delay before issuing the response
                // wait(100,SC_NS);
                // ---------------------------------------------------------
                // Sends the response using only one chunk 
                // ---------------------------------------------------------
                // Set the members of the response group stucture
                m_resp[3].SResp          =     OCP_SRESP_DVA;      // Response is VALID
                m_resp[3].SDataPtr       =     data;               // pass a pointer to the slave data
                m_resp[3].SRespInfo      =     1;                  // Info for the first chunk

                // Original OCP TL2 API:
                // assert(RespChunkLen[3] == 0);
                // RespChunkLen[3] = 20; RespChunkLast[3] = true;

                // New OCP TL2 API:
                assert(m_resp[3].DataLength == 0);
                m_resp[3].DataLength = 20; 
                m_resp[3].LastOfBurst = true;

                // try to ship this new respone out
                tryToSendResponse();
                m_req_num = 1;
                break;

            default:
                cout << "ERROR: slave reached unknown request number = " << m_req_num << endl;
        }

    }

    void releaseRequest_method()
    {
        // Original OCP TL2 API:
        // ret_value = ocp->putSCmdAccept();

        // New OCP TL2 API:
        ret_value = ocp->acceptRequest();

        assert(ret_value);
    }

    void send_response_method()
    {
        // Response Channel is free - try to send next response
        tryToSendResponse();
    }

    void tryToSendResponse()
    {
        if (m_resp[m_resp_num].DataLength == 0)
        {
            // Give up - our response is not ready yet
            return;
        }
        // cout << "Slave trying to send response " << m_resp_num << " at time " << sc_time_stamp() << endl;

        if ( ocp->sendOCPResponse(m_resp[m_resp_num]) )
        {
            // Success - response is on the channel
            // cout << "Slave was able to send response " << m_resp_num << " at time " << sc_time_stamp() << endl;
            m_resp[m_resp_num].DataLength = 0;
            m_resp_num++;
            if (m_resp_num == 4)
            {
                m_resp_num = 1;
                m_loops_complete++;
                if( m_loops_complete == LOOP_COUNT)
                {
                    cout << "Slave DONE" << endl;
                }
            }
        }
    }

    // ----------------------------------------------------------------------------
    // constuctor
    // ----------------------------------------------------------------------------
    slave(sc_module_name mod):
        sc_module(mod),
    ocp("Sale_Port")
    {
        SC_METHOD(get_request_method);
          sensitive << ocp.RequestStartEvent();
          dont_initialize();
        SC_METHOD(releaseRequest_method);
          sensitive << m_releaseRequestEvent;
          dont_initialize();
        SC_METHOD(send_response_method);
          sensitive << ocp.ResponseEndEvent();
          dont_initialize();

        // initialize variables
        for (int i=0;i<4;i++)
        {
            m_resp[i].reset();  
            // RespChunkLen[i] =0;  // No longer needed
            // RespChunkLast[i] =true;  // No longer needed
        }
        for (int i=0;i<100;i++)
        {
            data[i]=9000+i;
        }
        m_req_num=1;
        m_resp_num=1;
        m_loops_complete=0;
    }

    // ----------------------------------------------------------------------------
    // Internal class members 
    // ----------------------------------------------------------------------------
    private:
    // slave memory 
    unsigned int data[100];

    // OCP Request/Response structures
    OCPTL2RequestGrp<unsigned int, unsigned int> m_req;
    OCPTL2ResponseGrp<unsigned int> m_resp[4];
    // unsigned int RespChunkLen[4];  // No longer needed
    // bool RespChunkLast[4];  // No longer needed
    sc_event m_releaseRequestEvent;
    sc_event m_newResponseAvailableEvent;
    // bool release_immediately;  // No longer needed
    bool ret_value;
    // unsigned int ReqChunkLen;  // No longer needed
    // bool ReqChunkLast;  // No longer needed
    int m_req_num;
    int m_resp_num;
    int m_loops_complete;
    
};

// ----------------------------------------------------------------------------
// MAIN PROGRAM
// ----------------------------------------------------------------------------

void readMapFromFile(const string &myFileName, MapStringType &myParamMap) 
{
    // read pairs of data from the passed file
    string leftside;
    string rightside;
    
    // (1) open the file
    ifstream inputfile(myFileName.c_str());
    assert( inputfile );

    // set the formatting
    inputfile.setf(std::ios::skipws);

    // Now read through all the pairs of values and add them to the passed map
    while ( inputfile ) {
        inputfile >> leftside;
        inputfile >> rightside;
        myParamMap.insert(std::make_pair(leftside,rightside));
    }

    // All done, close up
    inputfile.close();
}

  // Performs some basic checks on the parameter values
  // To be completed....
  bool ConfigChecking(OCPParameters* pcl) {
     bool configError = false;
      
     // 'Exact ThreadBusy' conditions (OCP 2.0 spec. page 49)
     if (pcl->sthreadbusy_exact==1 && !(pcl->cmdaccept == 0) )
     {
         // cout <<  " OCP parameter error: " << endl;
         // cout << "      sthreadbusy_exact==1 and cmdaccept parameter is not 0" ;
         configError = true;
     }
     if (pcl->sthreadbusy_exact==1 && pcl->datahandshake==1 && pcl->sdatathreadbusy==0 && !(pcl->dataaccept == 0) )
     {
         // cout <<  " OCP parameter error: " << endl;
         // cout << "      sthreadbusy_exact==1,datahandshake==1,sdatathreadbusy==0 and dataaccept parameter is not 0" ;
         configError = true;
     }
     if (pcl->sdatathreadbusy_exact==1 && !(pcl->dataaccept == 0) )
     {
         // cout <<  " OCP parameter error: " << endl;
         // cout << "      sdatathreadbusy_exact==1 and dataaccept parameter is not 0" ;
         configError = true;
     }
     if (pcl->mthreadbusy_exact==1 && !(pcl->respaccept == 0) )
     {
         // cout <<  " OCP parameter error: " << endl;
         // cout << "      mthreadbusy_exact==1 and respaccept parameter is not 0" ;
         configError = true;
     }
     if (pcl->sthreadbusy_exact==0 && pcl->cmdaccept== 0 && pcl->sthreadbusy==1 )
     {
         // cout <<  " OCP parameter error: " << endl;
         // cout << "      sthreadbusy_exact is set to 0, cmdaccept is set to 0, but sthreadbusy is set to 1" ;
         configError = true;
     }
     if (pcl->mthreadbusy_exact==0 && pcl->respaccept== 0 && pcl->mthreadbusy==1 )
     {
         // cout <<  " OCP parameter error: " << endl;
         // cout << "      mthreadbusy_exact is set to 0, respaccept is set to 0, but mthreadbusy is set to 1" ;
         configError = true;
     }

     return !configError;
      
  };



int sc_main(int, char*[])
{

    // Creates the OCP TL2 channel
    OCP_TL2_Channel<unsigned int, unsigned int>  ch0("ch0");

    // Set the OCP parameters for this channel
    MapStringType  ocpParamMap;
    readMapFromFile("ocpParams", ocpParamMap);
    ch0.setConfiguration(ocpParamMap);

    // performs some basic checks on the OCP parameter values
    // Old style OCP parameter class:
    // ParamCl<OCP_TL2_DataCl<unsigned int, unsigned int> > *pcl = ch0.GetParamCl();
    // New style OCP parameter class (no template):
    OCPParameters *pcl = ch0.GetParamCl(); 
    assert(ConfigChecking(pcl));

    // Creates masters and slaves 
    slave sl1("sl1");
    master ms1("ms1");

    // Connect masters and slaves using channels
    ms1.ocp(ch0);
    sl1.ocp(ch0);

    // Starts simulation 
    sc_start(END_TIME, SC_MS);

    return(0);
}
